"""
The FiftyOne Zoo.

Copyright 2017-2022, Voxel51, Inc.
voxel51.com
"""
import types

from .datasets import *
from .models import *

# This enables Sphinx refs to directly use paths imported here
__all__ = [
    k
    for k, v in globals().items()
    if not k.startswith("_") and not isinstance(v, types.ModuleType)
]
