import{a6 as t,w as a,r as s,R as o,L as c}from"./index.e25a78a1.js";const n=({})=>{const e=t(a);return s.exports.useLayoutEffect(()=>{e(null)},[]),o.createElement(c,null,"No dataset selected")},l=n;export{l as default};
